/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

// moved these from glquake.h so that we can more easily change them
typedef enum
{
	pt_static, pt_grav, pt_slowgrav, pt_fire, pt_explode, pt_explode2, pt_blob, pt_blob2
} ptype_t;


typedef struct particle_s
{
	// driver-usable fields
	vec3_t		org;
	float		color;

	// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t		vel;
	float		ramp;
	float		die;
	ptype_t		type;
	float		alpha;
	float		alphadec;
} particle_t;


#define PARTICLE_BATCH_SIZE		2048
#define PARTICLE_EXTRA_SIZE		1024

// we don't expect these to ever be exceeded but we allow it if required
#define PARTICLE_TYPE_BATCH_SIZE	64
#define PARTICLE_TYPE_EXTRA_SIZE	32

int		ramp1[8] = {0x6f, 0x6d, 0x6b, 0x69, 0x67, 0x65, 0x63, 0x61};
int		ramp2[8] = {0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x68, 0x66};
int		ramp3[8] = {0x6d, 0x6b, 6, 5, 4, 3};

particle_t *free_particles;
particle_type_t *active_particle_types, *free_particle_types;
vec3_t r_pright, r_pup, r_ppn;

int r_numparticles = 0;

gltexture_t *particletexture, *particletexture1, *particletexture2, *particletexture3, *particletexture4; // johnfitz
float texturescalefactor; // johnfitz -- compensate for apparent size of different particle textures
float texcoordfactor;

cvar_t	r_particles = {"r_particles", "1", true}; // johnfitz

/*
===============
R_ParticleTextureLookup -- johnfitz -- generate nice antialiased 32x32 circle for particles
===============
*/
int R_ParticleTextureLookup (int x, int y, int sharpness)
{
	int r; // distance from point x,y to circle origin, squared
	int a; // alpha value to return

	x -= 16;
	y -= 16;
	r = x * x + y * y;
	r = r > 255 ? 255 : r;
	a = sharpness * (255 - r);
	a = min(a,255);
	return a;
}


/*
===============
R_InitParticleTextures -- johnfitz -- rewritten
===============
*/
byte particle4_data_src[] =
{
	0, 0, 1, 8, 8, 1, 0, 0,
	0, 8, 64, 128, 128, 64, 8, 0,
	1, 64, 192, 224, 224, 192, 64, 1,
	8, 128, 224, 255, 255, 224, 128, 8,
	8, 128, 224, 255, 255, 224, 128, 8,
	1, 64, 192, 224, 224, 192, 64, 1,
	0, 8, 64, 128, 128, 64, 8, 0,
	0, 0, 1, 8, 8, 1, 0, 0
};


void R_InitParticleTextures (void)
{
	// mh - this is eeeeeeeeeevil.
	// 8x8 is perfect for these and will perform much better too.  until then we'll just mipmap the bastards
	int			x, y, z;
	static byte	particle1_data[64*64*4];
	static byte	particle2_data[2*2*4];
	static byte	particle3_data[64*64*4];
	static byte	particle4_data[8*8*4];
	byte		*dst;

	// particle texture 1 -- circle
	dst = particle1_data;

	for (x = 0; x < 64; x++)
	{
		for (y = 0; y < 64; y++)
		{
			*dst++ = 255;
			*dst++ = 255;
			*dst++ = 255;
			*dst++ = R_ParticleTextureLookup (x, y, 8);
		}
	}

	particletexture1 = TM_LoadImage (NULL, "particle1", 64, 64, SRC_RGBA, particle1_data, "", (src_offset_t) particle1_data, TEXPREF_PERSIST | TEXPREF_ALPHA | TEXPREF_LINEAR);

	// particle texture 2 -- square
	dst = particle2_data;

	for (x = 0; x < 2; x++)
	{
		for (y = 0; y < 2; y++)
		{
			*dst++ = 255;
			*dst++ = 255;
			*dst++ = 255;
			*dst++ = x || y ? 0 : 255;
		}
	}

	particletexture2 = TM_LoadImage (NULL, "particle2", 2, 2, SRC_RGBA, particle2_data, "", (src_offset_t) particle2_data, TEXPREF_PERSIST | TEXPREF_ALPHA | TEXPREF_NEAREST);

	// particle texture 3 -- blob
	dst = particle3_data;

	for (x = 0; x < 64; x++)
	{
		for (y = 0; y < 64; y++)
		{
			*dst++ = 255;
			*dst++ = 255;
			*dst++ = 255;
			*dst++ = R_ParticleTextureLookup (x, y, 2);
		}
	}

	particletexture3 = TM_LoadImage (NULL, "particle3", 64, 64, SRC_RGBA, particle3_data, "", (src_offset_t) particle3_data, TEXPREF_PERSIST | TEXPREF_ALPHA | TEXPREF_LINEAR);

	// particle texture 4 -- low res circle
	dst = particle4_data;

	for (x = 0, z = 0; x < 8; x++)
	{
		for (y = 0; y < 8; y++, z++)
		{
			*dst++ = 255;
			*dst++ = 255;
			*dst++ = 255;
			*dst++ = particle4_data_src[z];
		}
	}

	particletexture4 = TM_LoadImage (NULL, "particle4", 8, 8, SRC_RGBA, particle4_data, "", (src_offset_t) particle4_data, TEXPREF_PERSIST | TEXPREF_ALPHA | TEXPREF_LINEAR);

	// set default
	particletexture = particletexture1;
	texturescalefactor = 1.27;
	texcoordfactor = 0.5f;
}


/*
===============
R_SetParticleTexture_f -- johnfitz
===============
*/
void R_SetParticleTexture_f (void)
{
	texcoordfactor = 0.5f;

	switch ((int) (r_particles.value))
	{
	case 1:
		particletexture = particletexture1;
		texturescalefactor = 1.27;
		break;
	case 2:
		particletexture = particletexture2;
		texturescalefactor = 1.0;
		break;
	case 3:
		particletexture = particletexture3;
		texturescalefactor = 1.5;
		break;
	case 4:
	default:
		particletexture = particletexture4;
		texturescalefactor = 1.5f;
		texcoordfactor = 1.0f;
		break;
	}
}


/*
===============
R_InitParticles
===============
*/
void R_InitParticles (void)
{
	Cvar_RegisterVariable (&r_particles, R_SetParticleTexture_f); // johnfitz
	R_InitParticleTextures (); // johnfitz
}


particle_t *R_NewParticle (particle_type_t *pt)
{
	particle_t	*p;
	int i;

	// with pointfiles we can have loads of particles, so set this as an absolute upper limit to prevent memory running out
	if (r_numparticles >= 65536)
		return NULL;
	else if (free_particles)
	{
		// just take from the free list
		p = free_particles;
		free_particles = p->next;

		p->next = pt->particles;
		pt->particles = p;
		pt->numparticles++;

		// set some default properties
		p->alpha = 1;
		p->alphadec = 0;

		return p;
	}
	else
	{
		// alloc some more free particles
		free_particles = (particle_t *) Hunk_AllocName (PARTICLE_EXTRA_SIZE * sizeof (particle_t), "particles");

		// link them up
		for (i = 1; i < PARTICLE_EXTRA_SIZE; i++)
		{
			free_particles[i - 1].next = &free_particles[i];
			free_particles[i].next = NULL;
		}

		// finish the link
		r_numparticles += PARTICLE_EXTRA_SIZE;

		// call recursively to return the first new free particle
		return R_NewParticle (pt);
	}
}


particle_type_t *R_NewParticleType (vec3_t spawnorg)
{
	particle_type_t *pt;
	int i;

	if (free_particle_types)
	{
		// just take from the free list
		pt = free_particle_types;
		free_particle_types = pt->next;

		// no particles yet
		pt->particles = NULL;
		pt->numparticles = 0;

		// copy across origin
		VectorCopy (spawnorg, pt->spawnorg);

		// link it in
		pt->next = active_particle_types;
		active_particle_types = pt;

		// done
		return pt;
	}

	// alloc some more free particles
	free_particle_types = (particle_type_t *) Hunk_Alloc (PARTICLE_TYPE_EXTRA_SIZE * sizeof (particle_type_t));

	// link them up
	for (i = 1; i < PARTICLE_TYPE_EXTRA_SIZE; i++)
	{
		free_particle_types[i - 1].next = &free_particle_types[i];
		free_particle_types[i].next = NULL;
	}

	// call recursively to return the first new free particle type
	return R_NewParticleType (spawnorg);
}


/*
===============
R_EntityParticles
===============
*/
#define NUMVERTEXNORMALS	162
extern	float	r_avertexnormals[NUMVERTEXNORMALS][3];
vec3_t	avelocities[NUMVERTEXNORMALS];
float	beamlength = 16;
vec3_t	avelocity = {23, 7, 3};
float	partstep = 0.01;
float	timescale = 0.01;

void R_EntityParticles (entity_t *ent)
{
	int			count;
	int			i;
	particle_t	*p;
	particle_type_t *pt;
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	vec3_t		forward;
	float		dist;

	dist = 64;
	count = 50;

	if (!avelocities[0][0])
	{
		for (i = 0; i < NUMVERTEXNORMALS * 3; i++)
			avelocities[0][i] = (rand() & 255) * 0.01;
	}

	pt = R_NewParticleType (ent->origin);

	for (i = 0; i < NUMVERTEXNORMALS; i++)
	{
		angle = cl.time * avelocities[i][0];
		sy = sin (angle);
		cy = cos (angle);
		angle = cl.time * avelocities[i][1];
		sp = sin (angle);
		cp = cos (angle);
		angle = cl.time * avelocities[i][2];
		sr = sin (angle);
		cr = cos (angle);

		forward[0] = cp * cy;
		forward[1] = cp * sy;
		forward[2] = -sp;

		if (!(p = R_NewParticle (pt)))
			return;

		p->die = cl.time + 0.01;
		p->color = 0x6f;
		p->type = pt_explode;

		p->org[0] = ent->origin[0] + r_avertexnormals[i][0] * dist + forward[0] * beamlength;
		p->org[1] = ent->origin[1] + r_avertexnormals[i][1] * dist + forward[1] * beamlength;
		p->org[2] = ent->origin[2] + r_avertexnormals[i][2] * dist + forward[2] * beamlength;
	}
}

/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	int		i;

	free_particles = (particle_t *) Hunk_AllocName (PARTICLE_BATCH_SIZE * sizeof (particle_t), "particles");

	for (i = 1; i < PARTICLE_BATCH_SIZE; i++)
	{
		free_particles[i - 1].next = &free_particles[i];
		free_particles[i].next = NULL;
	}

	// particle type chains
	free_particle_types = (particle_type_t *) Hunk_Alloc (PARTICLE_TYPE_BATCH_SIZE * sizeof (particle_type_t));
	active_particle_types = NULL;

	for (i = 1; i < PARTICLE_TYPE_BATCH_SIZE; i++)
	{
		free_particle_types[i - 1].next = &free_particle_types[i];
		free_particle_types[i].next = NULL;
	}

	r_numparticles = PARTICLE_BATCH_SIZE;
}


/*
===============
R_ReadPointFile_f
===============
*/
void R_ReadPointFile_f (void)
{
	FILE	*f;
	vec3_t	org;
	int		r;
	int		c;
	particle_t	*p;
	particle_type_t *pt;
	char	name[MAX_OSPATH];

	sprintf (name, "maps/%s.pts", sv.name);

	COM_FOpenFile (name, &f);

	if (!f)
	{
		Con_Printf ("couldn't open %s\n", name);
		return;
	}

	Con_Printf ("Reading %s...\n", name);
	c = 0;

	pt = R_NewParticleType (vec3_origin);

	for (;;)
	{
		r = fscanf (f, "%f %f %f\n", &org[0], &org[1], &org[2]);

		if (r != 3)
			break;

		c++;

		if (!(p = R_NewParticle (pt)))
		{
			Con_Printf ("Not enough free particles\n");
			break;
		}

		p->die = 99999;
		p->color = (-c) & 15;
		p->type = pt_static;
		VectorCopy (vec3_origin, p->vel);
		VectorCopy (org, p->org);
	}

	fclose (f);
	Con_Printf ("%i points read\n", c);
}

/*
===============
R_ParseParticleEffect

Parse an effect out of the server message
===============
*/
void R_ParseParticleEffect (void)
{
	vec3_t		org, dir;
	int			i, count, msgcount, color;

	for (i = 0; i < 3; i++)
		org[i] = MSG_ReadCoord (cl.protocol);	// gb, PROTOCOL_RMQ

	for (i = 0; i < 3; i++)
		dir[i] = MSG_ReadChar () * (1.0 / 16);

	msgcount = MSG_ReadByte ();
	color = MSG_ReadByte ();

	if (msgcount == 255)
		count = 1024;
	else
		count = msgcount;

	R_RunParticleEffect (org, dir, color, count);
}

/*
===============
R_ParticleExplosion
===============
*/
void R_ParticleExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;
	particle_type_t *pt = R_NewParticleType (org);

	for (i = 0; i < 1024; i++)
	{
		if (!(p = R_NewParticle (pt)))
			return;

		p->die = cl.time + 5;
		p->color = ramp1[0];
		p->ramp = rand() & 3;

		if (i & 1)
		{
			p->type = pt_explode;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand() % 32) - 16);
				p->vel[j] = (rand() % 512) - 256;
			}
		}
		else
		{
			p->type = pt_explode2;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand() % 32) - 16);
				p->vel[j] = (rand() % 512) - 256;
			}
		}
	}
}

/*
===============
R_ParticleExplosion2
===============
*/
void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength)
{
	int			i, j;
	particle_t	*p;
	int			colorMod = 0;
	particle_type_t *pt = R_NewParticleType (org);

	for (i = 0; i < 512; i++)
	{
		if (!(p = R_NewParticle (pt)))
			return;

		p->die = cl.time + 0.3;
		p->color = colorStart + (colorMod % colorLength);
		colorMod++;

		p->type = pt_blob;

		for (j = 0; j < 3; j++)
		{
			p->org[j] = org[j] + ((rand() % 32) - 16);
			p->vel[j] = (rand() % 512) - 256;
		}
	}
}

/*
===============
R_BlobExplosion
===============
*/
void R_BlobExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;
	particle_type_t *pt = R_NewParticleType (org);

	for (i = 0; i < 1024; i++)
	{
		if (!(p = R_NewParticle (pt)))
			return;

		p->die = cl.time + 1 + (rand() & 8) * 0.05;

		if (i & 1)
		{
			p->type = pt_blob;
			p->color = 66 + rand() % 6;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand() % 32) - 16);
				p->vel[j] = (rand() % 512) - 256;
			}
		}
		else
		{
			p->type = pt_blob2;
			p->color = 150 + rand() % 6;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand() % 32) - 16);
				p->vel[j] = (rand() % 512) - 256;
			}
		}
	}
}

/*
===============
R_RunParticleEffect
===============
*/
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	particle_t	*p;
	particle_type_t *pt = R_NewParticleType (org);

	for (i = 0; i < count; i++)
	{
		if (!(p = R_NewParticle (pt)))
			return;

		if (count == 1024)
		{
			// rocket explosion
			p->die = cl.time + 5;
			p->color = ramp1[0];
			p->ramp = rand() & 3;

			if (i & 1)
			{
				p->type = pt_explode;

				for (j = 0; j < 3; j++)
				{
					p->org[j] = org[j] + ((rand() % 32) - 16);
					p->vel[j] = (rand() % 512) - 256;
				}
			}
			else
			{
				p->type = pt_explode2;

				for (j = 0; j < 3; j++)
				{
					p->org[j] = org[j] + ((rand() % 32) - 16);
					p->vel[j] = (rand() % 512) - 256;
				}
			}
		}
		else
		{
			p->die = cl.time + 0.1 * (rand() % 5);
			p->color = (color&~7) + (rand() & 7);
			p->type = pt_slowgrav;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand() & 15) - 8);
				p->vel[j] = dir[j] * 15; // + (rand()%300)-150;
			}
		}
	}
}

/*
===============
R_LavaSplash
===============
*/
void R_LavaSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;
	particle_type_t *pt = R_NewParticleType (org);

	for (i = -16; i < 16; i++)
	{
		for (j = -16; j < 16; j++)
		{
			for (k = 0; k < 1; k++)
			{
				if (!(p = R_NewParticle (pt)))
					return;

				p->die = cl.time + 2 + (rand() & 31) * 0.02;
				p->color = 224 + (rand() & 7);
				p->type = pt_slowgrav;

				dir[0] = j * 8 + (rand() & 7);
				dir[1] = i * 8 + (rand() & 7);
				dir[2] = 256;

				p->org[0] = org[0] + dir[0];
				p->org[1] = org[1] + dir[1];
				p->org[2] = org[2] + (rand() & 63);

				VectorNormalize (dir);
				vel = 50 + (rand() & 63);
				VectorScale (dir, vel, p->vel);
			}
		}
	}
}

/*
===============
R_TeleportSplash
===============
*/
void R_TeleportSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;
	particle_type_t *pt = R_NewParticleType (org);

	for (i = -16; i < 16; i += 4)
	{
		for (j = -16; j < 16; j += 4)
		{
			for (k = -24; k < 32; k += 4)
			{
				if (!(p = R_NewParticle (pt)))
					return;

				p->die = cl.time + 0.2 + (rand() & 7) * 0.02;
				p->color = 7 + (rand() & 7);
				p->type = pt_slowgrav;

				dir[0] = j * 8;
				dir[1] = i * 8;
				dir[2] = k * 8;

				p->org[0] = org[0] + i + (rand() & 3);
				p->org[1] = org[1] + j + (rand() & 3);
				p->org[2] = org[2] + k + (rand() & 3);

				VectorNormalize (dir);
				vel = 50 + (rand() & 63);
				VectorScale (dir, vel, p->vel);
			}
		}
	}
}


void R_SpawnSmoke (vec3_t org, float jitter, float time)
{
}


#define Q_Random(MIN,MAX) ((rand() & 32767) * (((MAX) - (MIN)) * (1.0f / 32767.0f)) + (MIN))

void R_SpawnRainParticle (particle_type_t *pt, vec3_t mins, vec3_t maxs, int color, vec3_t vel, float time, float z)
{
	particle_t *p = R_NewParticle (pt);

	if (!p) return;

	p->org[0] = Q_Random (mins[0], maxs[0]);
	p->org[1] = Q_Random (mins[1], maxs[1]);
	p->org[2] = z;

	p->vel[0] = vel[0];
	p->vel[1] = vel[1];
	p->vel[2] = vel[2];

	p->color = color;
	p->type = pt_grav;

	p->die = cl.time + time;
}


void R_ParticleRain (vec3_t mins, vec3_t maxs, vec3_t dir, int count, int colorbase, int type)
{
	vec3_t		vel;
	float		t, z;

	particle_type_t *pt = R_NewParticleType (mins);

	// increase the number of particles
	count *= 2;

	if (maxs[0] <= mins[0]) {t = mins[0]; mins[0] = maxs[0]; maxs[0] = t;}
	if (maxs[1] <= mins[1]) {t = mins[1]; mins[1] = maxs[1]; maxs[1] = t;}
	if (maxs[2] <= mins[2]) {t = mins[2]; mins[2] = maxs[2]; maxs[2] = t;}

	if (dir[2] < 0) // falling
	{
		t = (maxs[2] - mins[2]) / -dir[2];
		z = maxs[2];
	}
	else // rising??
	{
		t = (maxs[2] - mins[2]) / dir[2];
		z = mins[2];
	}

	if (t < 0 || t > 2) // sanity check
		t = 2;

	switch (type)
	{
	case 0:
		while (count--)
		{
			vel[0] = dir[0] + Q_Random (-16, 16);
			vel[1] = dir[1] + Q_Random (-16, 16);
			vel[2] = dir[2] + Q_Random (-32, 32);

			R_SpawnRainParticle (pt, mins, maxs, colorbase + (rand () & 3), vel, t, z);
		}

		break;

	case 1:
		while (count--)
		{
			vel[0] = dir[0] + Q_Random (-16, 16);
			vel[1] = dir[1] + Q_Random (-16, 16);
			vel[2] = dir[2] + Q_Random (-32, 32);

			R_SpawnRainParticle (pt, mins, maxs, colorbase + (rand () & 3), vel, t, z);
		}

		break;

	default:
		Con_DPrintf ("R_ParticleRain: unknown type %i (0 = rain, 1 = snow)\n", type);
	}
}


void MakeNormalVectors (vec3_t forward, vec3_t right, vec3_t up)
{
	float		d;

	// this rotate and negate guarantees a vector
	// not colinear with the original
	right[1] = -forward[0];
	right[2] = forward[1];
	right[0] = forward[2];

	d = DotProduct (right, forward);
	VectorMA (right, -d, forward, right);
	VectorNormalize (right);
	CrossProduct (right, forward, up);
}


float crand (void) {return (rand () & 32767) * (2.0 / 32767) - 1;}
float frand (void) {return (rand () & 32767) * (1.0 / 32767);}


void R_RailTrail (vec3_t start, vec3_t end, byte corecolor, byte spiralcolor)
{
	vec3_t		move;
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	float		dec;
	vec3_t		right, up;
	int			i;
	float		d, c, s;
	vec3_t		dir;
	particle_type_t *pt = R_NewParticleType (start);

	VectorCopy (start, move);
	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	MakeNormalVectors (vec, right, up);

	for (i = 0; i < len; i += 2)
	{
		if (!(p = R_NewParticle (pt)))
			return;

		d = i * 0.1;
		c = cos (d);
		s = sin (d);

		VectorScale (right, c, dir);
		VectorMA (dir, s, up, dir);

		p->die = cl.time + 666;
		p->color = spiralcolor + (rand () & 7);

		for (j = 0; j < 3; j++)
		{
			p->org[j] = move[j] + dir[j] * 3;
			p->vel[j] = dir[j] * 6;
		}

		VectorAdd (move, vec, move);
		p->type = pt_static;

		p->alpha = 1.25f;
		p->alphadec = 1.0 / (1 + frand () * 0.2);
	}

	dec = 0.75;
	VectorScale (vec, dec, vec);
	VectorCopy (start, move);

	while (len > 0)
	{
		len -= dec;

		if (!(p = R_NewParticle (pt)))
			return;

		p->die = cl.time + 666;
		p->color = corecolor + (rand () & 7);

		for (j = 0; j < 3; j++)
		{
			p->org[j] = move[j] + crand ();
			p->vel[j] = crand ();
		}

		VectorAdd (move, vec, move);
		p->type = pt_static;

		p->alpha = 1.5f;
		p->alphadec = 1.0 / (0.6 + frand() * 0.2);
	}
}


/*
===============
R_RocketTrail

FIXME -- rename function and use #defined types instead of numbers
===============
*/
void R_RocketTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;
	particle_type_t *pt = R_NewParticleType (start);

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 3;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len > 0)
	{
		len -= dec;

		if (!(p = R_NewParticle (pt)))
			return;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		switch (type)
		{
		case 0:	// rocket trail
			p->ramp = (rand() & 3);
			p->color = ramp3[(int) p->ramp];
			p->type = pt_fire;

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand() % 6) - 3);

			break;

		case 1:	// smoke smoke
			p->ramp = (rand() & 3) + 2;
			p->color = ramp3[(int) p->ramp];
			p->type = pt_fire;

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand() % 6) - 3);

			break;

		case 2:	// blood
			p->type = pt_grav;
			p->color = 67 + (rand() & 3);

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand() % 6) - 3);

			break;

		case 3:
		case 5:	// tracer
			p->die = cl.time + 0.5;
			p->type = pt_static;

			if (type == 3)
				p->color = 52 + ((tracercount & 4) << 1);
			else
				p->color = 230 + ((tracercount & 4) << 1);

			tracercount++;

			VectorCopy (start, p->org);

			if (tracercount & 1)
			{
				p->vel[0] = 30 * vec[1];
				p->vel[1] = 30 * -vec[0];
			}
			else
			{
				p->vel[0] = 30 * -vec[1];
				p->vel[1] = 30 * vec[0];
			}

			break;

		case 4:	// slight blood
			p->type = pt_grav;
			p->color = 67 + (rand() & 3);

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand() % 6) - 3);

			len -= 3;
			break;

		case 6:	// voor trail
			p->color = 9 * 16 + 8 + (rand() & 3);
			p->type = pt_static;
			p->die = cl.time + 0.3;

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand() & 15) - 8);

			break;
		}


		VectorAdd (start, vec, start);
	}
}

/*
===============
CL_RunParticles -- johnfitz -- all the particle behavior, separated from R_DrawParticles
===============
*/
void CL_RunParticlesForType (particle_type_t *pt)
{
	particle_t		*p, *kill;
	int				i;
	float			time1, time2, time3, dvel, frametime, grav;
	extern	cvar_t	sv_gravity;

	frametime = fabs (cl.ctime - cl.oldtime);
	time3 = frametime * 15;
	time2 = frametime * 10;
	time1 = frametime * 5;
	grav = frametime * sv_gravity.value * 0.05;
	dvel = 4 * frametime;

	for (;;)
	{
		kill = pt->particles;

		if (kill && kill->die < cl.time)
		{
			pt->particles = kill->next;
			kill->next = free_particles;
			free_particles = kill;
			r_numparticles--;
			pt->numparticles--;
			continue;
		}

		break;
	}

	for (p = pt->particles; p; p = p->next)
	{
		for (;;)
		{
			kill = p->next;

			if (kill && kill->die < cl.time)
			{
				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;
				r_numparticles--;
				pt->numparticles--;
				continue;
			}

			break;
		}

		p->org[0] += p->vel[0] * frametime;
		p->org[1] += p->vel[1] * frametime;
		p->org[2] += p->vel[2] * frametime;

		p->alpha -= p->alphadec * frametime;

		if (p->alpha <= 0) p->die = -1;

		switch (p->type)
		{
		case pt_static:
			break;

		case pt_fire:
			p->ramp += time1;

			if (p->ramp >= 6)
				p->die = -1;
			else
				p->color = ramp3[(int) p->ramp];

			p->vel[2] += grav;
			break;

		case pt_explode:
			p->ramp += time2;

			if (p->ramp >= 8)
				p->die = -1;
			else
				p->color = ramp1[(int) p->ramp];

			for (i = 0; i < 3; i++)
				p->vel[i] += p->vel[i] * dvel;

			p->vel[2] -= grav;
			break;

		case pt_explode2:
			p->ramp += time3;

			if (p->ramp >= 8)
				p->die = -1;
			else
				p->color = ramp2[(int) p->ramp];

			for (i = 0; i < 3; i++)
				p->vel[i] -= p->vel[i] * frametime;

			p->vel[2] -= grav;
			break;

		case pt_blob:
			for (i = 0; i < 3; i++)
				p->vel[i] += p->vel[i] * dvel;

			p->vel[2] -= grav;
			break;

		case pt_blob2:

			for (i = 0; i < 2; i++)
				p->vel[i] -= p->vel[i] * dvel;

			p->vel[2] -= grav;
			break;

		case pt_grav:
		case pt_slowgrav:
			p->vel[2] -= grav;
			break;
		}
	}
}


void CL_RunParticles (void)
{
	particle_type_t	*pt, *kill;

	// remove from the head of the list
	for (;;)
	{
		kill = active_particle_types;

		if (kill && !kill->particles)
		{
			// return to the free list
			active_particle_types = kill->next;
			kill->next = free_particle_types;
			kill->numparticles = 0;
			free_particle_types = kill;

			continue;
		}

		break;
	}

	// no types currently active
	if (!active_particle_types) return;

	for (pt = active_particle_types; pt; pt = pt->next)
	{
		// remove from a mid-point in the list
		for (;;)
		{
			kill = pt->next;

			if (kill && !kill->particles)
			{
				pt->next = kill->next;
				kill->next = free_particle_types;
				kill->numparticles = 0;
				free_particle_types = kill;

				continue;
			}

			break;
		}

		CL_RunParticlesForType (pt);
	}
}


/*
===============
R_DrawParticles -- johnfitz -- moved all non-drawing code to CL_RunParticles
===============
*/
extern unsigned int r_quadindexbuffer;
extern unsigned int r_parttexcoordbuffer;


float up[3];
float right[3];

int r_particleframe = -1;

typedef struct r_partgs_s
{
	float xyz[3];

	union
	{
		byte rgba[4];
		int c;
	};
} r_partgs_t;


r_partgs_t *r_gsparticles;
int r_numgsparticles;
int r_maxgsparticles = (SCRATCHBUF_SIZE / sizeof (r_partgs_t));


void R_ParticlesBegin (void)
{
	// mh - should this be adjustable per particle???
	VectorScale (vup, 1.5, up);
	VectorScale (vright, 1.5, right);

	if (r_maxgsparticles > 8192)
		r_maxgsparticles = 8192;

	if (gl_arb_point_parameters)
	{
		float attenuations[3] = {0.01f, 0.0f, 0.01f};

		// ATI drivers back in the old days wouldn't tolerate this being set only once during startup so we set it every frame
		glEnable (GL_POINT_SMOOTH);

		// ATI drivers back in the old days reported the extension but didn't export glPointParameteri...
		qglPointParameterf (GL_POINT_SIZE_MIN_ARB, 2);
		qglPointParameterf (GL_POINT_SIZE_MAX_ARB, 50);
		qglPointParameterfv (GL_POINT_DISTANCE_ATTENUATION_ARB, attenuations);

		r_gsparticles = (r_partgs_t *) scratchbuf;
		r_numgsparticles = 0;

		GL_SetStreamSource (0, GLSTREAM_POSITION, 3, GL_FLOAT, sizeof (r_partgs_t), r_gsparticles->xyz);
		GL_SetStreamSource (0, GLSTREAM_COLOR, 4, GL_UNSIGNED_BYTE, sizeof (r_partgs_t), r_gsparticles->rgba);
		GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 0, GL_NONE, 0, NULL);
		GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
		GL_SetStreamSource (0, GLSTREAM_TEXCOORD2, 0, GL_NONE, 0, NULL);

		glDisable (GL_TEXTURE_2D);
		glPointSize (50);
	}
	else
	{
		GL_BindTexture (GL_TEXTURE0, particletexture);

		r_num_quads = 0;

		if (gl_arb_vertex_buffer_object)
		{
			glMatrixMode (GL_TEXTURE);
			glLoadIdentity ();
			glScalef (texcoordfactor, texcoordfactor, 1);
			glMatrixMode (GL_MODELVIEW);

			GL_SetStreamSource (0, GLSTREAM_POSITION, 3, GL_FLOAT, sizeof (r_defaultquad2_t), r_default_quads2->xyz);
			GL_SetStreamSource (0, GLSTREAM_COLOR, 4, GL_UNSIGNED_BYTE, sizeof (r_defaultquad2_t), r_default_quads2->color);
			GL_SetStreamSource (r_parttexcoordbuffer, GLSTREAM_TEXCOORD0, 2, GL_FLOAT, 0, (void *) 0);
		}
		else
		{
			GL_SetStreamSource (0, GLSTREAM_POSITION, 3, GL_FLOAT, sizeof (r_defaultquad_t), r_default_quads->xyz);
			GL_SetStreamSource (0, GLSTREAM_COLOR, 4, GL_UNSIGNED_BYTE, sizeof (r_defaultquad_t), r_default_quads->color);
			GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 2, GL_FLOAT, sizeof (r_defaultquad_t), r_default_quads->st);
		}

		GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_MODULATE);

		GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
		GL_SetStreamSource (0, GLSTREAM_TEXCOORD2, 0, GL_NONE, 0, NULL);
	}

	glDepthMask (GL_TRUE);
}


void R_ParticlesEnd (void)
{
	if (gl_arb_point_parameters)
	{
		if (r_numgsparticles)
			GL_DrawPrimitive (GL_POINTS, 0, r_numgsparticles);

		glDisable (GL_POINT_SMOOTH);
		glEnable (GL_TEXTURE_2D);
	}
	else
	{
		// draw anything left over
		if (r_num_quads)
		{
			if (gl_arb_vertex_buffer_object && r_quadindexbuffer)
				GL_SetIndices (r_quadindexbuffer, (void *) 0);
			else GL_SetIndices (0, r_quad_indexes);

			GL_DrawIndexedPrimitive (GL_TRIANGLES, r_num_quads * 6, r_num_quads * 4);

			r_num_quads = 0;
		}

		GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_REPLACE);

		if (gl_arb_vertex_buffer_object)
		{
			glMatrixMode (GL_TEXTURE);
			glLoadIdentity ();
			glMatrixMode (GL_MODELVIEW);

			GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 0, GL_NONE, 0, NULL);
		}
	}

	glDepthMask (GL_FALSE);
}


#define PARTICLE_VERTEX(i,vert) \
	rpq[i].xyz[0] = (vert)[0]; \
	rpq[i].xyz[1] = (vert)[1]; \
	rpq[i].xyz[2] = (vert)[2];

#define PARTICLE_VERTEX2(xyz,vert) \
	xyz[0] = (vert)[0]; \
	xyz[1] = (vert)[1]; \
	xyz[2] = (vert)[2];

#define PARTICLE_TEXCOORD(i,s,t) \
	rpq[i].st[0] = s; \
	rpq[i].st[1] = t;

void R_DrawParticlesForType (particle_type_t *pt)
{
	particle_t *p;

	if (gl_arb_point_parameters)
	{
		for (p = pt->particles; p; p = p->next)
		{
			// don't crash (255 is alpha)
			if (p->color < 0 || p->color > 254) continue;
			if (p->alpha < 0) continue;

			// runs faster
			if (R_CullSphere (p->org, 64, 15)) continue;

			if (r_numgsparticles == r_maxgsparticles)
			{
				GL_DrawPrimitive (GL_POINTS, 0, r_numgsparticles);
				r_gsparticles = (r_partgs_t *) scratchbuf;
				r_numgsparticles = 0;
			}

			VectorCopy (p->org, r_gsparticles->xyz);

			// set correct particle colour
			r_gsparticles->c = d_8to24table_rgba[(int) p->color];
			r_gsparticles->rgba[3] = BYTE_CLAMPF (p->alpha);

			r_gsparticles++;
			r_numgsparticles++;
		}
	}
	else
	{
		float scale;
		vec3_t p_up, p_right, p_upright; // johnfitz -- p_ vectors

		r_defaultquad_t *rpq;
		r_defaultquad2_t *rpq2;

		unsigned int rgba;

		rpq = &r_default_quads[r_num_quads * 4];
		rpq2 = &r_default_quads2[r_num_quads * 4];

		for (p = pt->particles; p; p = p->next, rpq += 4, rpq2 += 4, r_num_quads++)
		{
			// don't crash (255 is alpha)
			if (p->color < 0 || p->color > 254) continue;
			if (p->alpha < 0) continue;

			if (r_num_quads >= r_max_quads)
			{
				if (gl_arb_vertex_buffer_object && r_quadindexbuffer)
					GL_SetIndices (r_quadindexbuffer, (void *) 0);
				else GL_SetIndices (0, r_quad_indexes);

				GL_DrawIndexedPrimitive (GL_TRIANGLES, r_num_quads * 6, r_num_quads * 4);

				rpq = r_default_quads;
				rpq2 = r_default_quads2;
				r_num_quads = 0;
			}

			// hack a scale up to keep particles from disapearing
			scale = (p->org[0] - r_origin[0]) * vpn[0] +
					(p->org[1] - r_origin[1]) * vpn[1] +
					(p->org[2] - r_origin[2]) * vpn[2];

			// mh - changed scales so that we don't need to halve it
			if (scale < 20)
				scale = 0.5 + 0.04; // johnfitz -- added .08 to be consistent
			else scale = 0.5 + scale * 0.002;

			scale *= texturescalefactor; // johnfitz -- compensate for apparent size of different particle textures

			VectorMA (p->org, scale, up, p_up);
			VectorMA (p->org, scale, right, p_right);
			VectorMA (p_up, scale, right, p_upright);

			// set correct particle colour
			rgba = d_8to24table_rgba[(int) p->color];
			((byte *) &rgba)[3] = BYTE_CLAMPF (p->alpha);

			if (gl_arb_vertex_buffer_object)
			{
				rpq2[0].rgba = rpq2[1].rgba = rpq2[2].rgba = rpq2[3].rgba = rgba;

				PARTICLE_VERTEX2 (rpq2[0].xyz, p->org);
				PARTICLE_VERTEX2 (rpq2[1].xyz, p_up);
				PARTICLE_VERTEX2 (rpq2[2].xyz, p_upright);
				PARTICLE_VERTEX2 (rpq2[3].xyz, p_right);
			}
			else
			{
				rpq[0].rgba = rpq[1].rgba = rpq[2].rgba = rpq[3].rgba = rgba;

				PARTICLE_VERTEX (0, p->org);
				PARTICLE_TEXCOORD (0, 0, 0);

				PARTICLE_VERTEX (1, p_up);
				PARTICLE_TEXCOORD (1, texcoordfactor, 0);

				PARTICLE_VERTEX (2, p_upright);
				PARTICLE_TEXCOORD (2, texcoordfactor, texcoordfactor);

				PARTICLE_VERTEX (3, p_right);
				PARTICLE_TEXCOORD (3, 0, texcoordfactor);
			}
		}
	}
}


void R_AddParticlesToAlphaList (particle_type_t *pt);

void R_DrawParticles (void)
{
	particle_type_t *pt;
	extern	cvar_t	r_particles; // johnfitz

	if (!r_particles.value) return;
	if (!active_particle_types) return;

	for (pt = active_particle_types; pt; pt = pt->next)
	{
		if (!pt->particles) continue;

		R_AddParticlesToAlphaList (pt);
	}
}
